from schemas import AJson
from schemas import MA
from pytest import fixture, mark


@fixture
def ajson_basic():
    return AJson.model_validate_json(open('tests/data/analysis_json_basic.json', 'r').read())


@fixture
def ajson_full():
    return AJson.model_validate_json(open('tests/data/analysis_json_full.json', 'r').read())

@mark.skip("no input data")
def test_ajson_basic(ajson_basic: AJson):
    assert ajson_basic.sourceBin is None
    assert isinstance(ajson_basic.mans[0], MA)

@mark.skip("no input data")
def test_ajson_full(ajson_full: AJson):
    assert ajson_full.mans[0].mdef['info']['short_name'] == 'sLoop'
    assert ajson_full.tStart == ajson_full.flown[0].time    

    #ajson_full.mans[0].flown['data'][0]['t']

#def test_sa_parse_ajson_basic(ajson_basic: AJson):
#    sa = ScheduleAnalysis.parse_dict(ajson_basic)
#    assert isinstance(sa, ScheduleAnalysis)
#
#def test_sa_parse_ajson_full(ajson_full: AJson):
#    sa = ScheduleAnalysis.parse_ajson(ajson_full)
#    assert isinstance(sa, ScheduleAnalysis)
#    assert len(sa) == 17
#    assert isinstance(sa[0].mdef, ManDef)
    


from enum import Enum
from pydantic_core import CoreSchema, core_schema
from pydantic import GetCoreSchemaHandler
from typing import Any

##https://github.com/pydantic/pydantic/discussions/6466#discussioncomment-8219585
class EnumStr(float, Enum):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.validate,
            core_schema.any_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: x.name
            ),
        )

    @classmethod
    def validate(cls, v: Any):
        if isinstance(v, cls):
            return v
        elif isinstance(v, str):
            try:
                return cls[v]
            except KeyError:
                return None
#                raise ValueError(f"Invalid {cls.__name__}: {v}")
        else:
            raise ValueError(f"Unexpected type: {type(v)}")


from schemas import ajson
from pathlib import Path


def test_validate():
    _ajson = ajson.AJson.model_validate_json((Path(__file__).parent / "data/flight.ajson").open().read())

    assert isinstance(_ajson, ajson.AJson)



from pydantic import BaseModel

from schemas.sinfo import ScheduleInfo
from schemas.positioning import Direction
from schemas.maninfo import ManInfo

class MDef(BaseModel):
    info: ManInfo
    mps: dict
    eds: dict
    box: dict


class MOption(BaseModel):
    options: list[MDef]
    active: int = 0

    @property
    def uid(self):
        return self.options[0].info.short_name

    @property
    def info(self):
        return self.options[self.active].info

    @property
    def mps(self):
        return self.options[self.active].mps

    @property
    def eds(self):
        return self.options[self.active].eds

    def __iter__(self):
        for mdef in self.options:
            yield mdef


class DirectionDefinition(BaseModel):
    manid: int
    direction: Direction


class SDefFile(BaseModel):
    category: str
    schedule: str
    fa_version: str
    mdefs: dict[str, dict | list[dict]]

    @property
    def sinfo(self):
        return ScheduleInfo(self.category, self.schedule)




import os
from json import load


def validate_json(file: dict|str|os.PathLike) -> dict:
    if isinstance(file, dict):
        return file
    elif isinstance(file, str) or isinstance(file, os.PathLike):
        with open(file, 'r') as f:
            return load(f)
    else:
        raise ValueError("expected a dict, str or os.PathLike")
    



from __future__ import annotations
from pydantic import BaseModel
from schemas import fcj
import pandas as pd
from datetime import datetime
from schemas.utils.files import validate_json
from schemas.ma import MA
from schemas.sinfo import ScheduleInfo
from packaging.version import InvalidVersion, Version


class AJson(BaseModel):
    origin: fcj.Origin | None = None
    isComp: bool
    sourceBin: str | None = None
    sourceFCJ: str | None = None
    bootTime: datetime | None = None
    mans: list[MA]

    def schedule(self):
        schedules = [man.schedule for man in self.mans]
        if all([s == schedules[0] for s in schedules[1:]]):
            return schedules[0].fcj_to_pfc()
        else:
            return ScheduleInfo.mixed()

    def all_versions(self):
        versions = set()
        for man in self.mans:
            versions |= set(man.history.keys())
        return list(versions)

    def all_valid_versions(self):
        valid_versions = []
        for version in self.all_versions():
            try:
                Version(version)
                valid_versions.append(version)
            except InvalidVersion:
                pass
        return valid_versions

    def get_scores(self, version: str, props: fcj.ScoreProperties=None, group="total"):
        if props is None:
            props = fcj.ScoreProperties()
        scores = {}
        for man in self.mans:
            if version in man.history:
                scores[man.name] = man.history[version].get_score(props).__dict__[group]
        return pd.Series(scores, name=version)

    def create_score_df(
        self, props: fcj.ScoreProperties, group="total", version: str = "All"
    ):
        versions = self.all_versions() if version == "All" else [version]
        return pd.concat([self.get_scores(ver, props, group) for ver in versions], axis=1)

    def check_version(self, version: str):
        version = version[1:] if version.startswith('v') else version
        return all([man.history is not None and version in man.history.keys() for man in self.mans])

    @staticmethod
    def parse_json(json: dict|str):
        return AJson.model_validate(validate_json(json))



from schemas import fcj
from pathlib import Path


def test_validate():
    _fcj = fcj.FCJ.model_validate_json((Path(__file__).parent / "data/fc_json.json").open().read())

    assert isinstance(_fcj, fcj.FCJ)


from schemas.maninfo import ManInfo
from pydantic import BaseModel
from typing import Any


class PE(BaseModel):
    kind: str
    args: list[Any]
    kwargs: dict[str, Any]
    centred: bool = False

    @staticmethod
    def build(kind: str, args: list, kwargs: dict, centred: bool = False):
        return PE(kind=kind, args=args, kwargs=kwargs, centred=centred)

    def __str__(self):
        pe = f"{self.kind}({','.join(self.args)}, {','.join([f'{k}={v}' for k,v in self.kwargs.items()])})"
        if self.centred:
            pe = f"centred({pe})"
        return pe


def line(*args, **kwargs):
    return PE.build("line", args, kwargs)


def roll(*args, **kwargs):
    return PE.build("roll", args, kwargs)


def loop(*args, **kwargs):
    return PE.build("loop", args, kwargs)


def snap(*args, **kwargs):
    return PE.build("snap", args, kwargs)


def spin(*args, **kwargs):
    return PE.build("spin", args, kwargs)


def stallturn(*args, **kwargs):
    return PE.build("stallturn", args, kwargs)


def centred(_pe: PE) -> PE:
    return _pe.model_copy(update=dict(centred=True))


class Figure(BaseModel):
    info: ManInfo
    elements: list[PE|int]
    ndmps: dict[str, float | int | list]
    relax_back: bool = False


def figure(
    info: ManInfo, elements: list[PE], relax_back: bool = False, **kwargs
) -> Figure:
    return Figure(info=info, elements=elements, ndmps=kwargs, relax_back=relax_back)


class Option(BaseModel):
    figures: list[Figure]

    @property
    def info(self):
        return self.figures[0].info
    

def option(figures: list[Figure]) -> Option:
    return Option(figures=figures)


class Sequence(BaseModel):
    name: str
    rules: str
    figures: list[Figure | Option]

def sequence(name: str, rules: str, figures: list[Figure | Option]) -> Sequence:
    return Sequence(name=name, rules=rules, figures=figures)


from .utils import validate_json, EnumStr 
from .positioning import (
    Heading,
    Direction,
    Orientation,
    Height,
    Position,
    BoxLocation,
    boxlocationmaker,
)
from .maninfo import ManInfo, maninfomaker
from .sinfo import ScheduleInfo, ManDetails
from .sdef import MDef, MOption, DirectionDefinition, SDefFile
from .ma import MA
from .ajson import AJson
from .aresti import (
    Sequence,
    sequence,
    Figure,
    figure,
    Option,
    option,
    PE,
    line,
    loop,
    option,
    roll,
    snap,
    spin,
    stallturn,
    centred
)


from __future__ import annotations
from dataclasses import dataclass
from typing import Literal
from schemas import Direction

fcj_categories = {
    "F3A FAI": "f3a",
    "F3A": "f3a",
    "US AMA": "nsrca",
    "F3A UK": "f3auk",
    "F3A US": "nsrca",
    "IMAC": "imac",
}

fcj_schedules = {
    "P23": "p23",
    "F23": "f23",
    "P25": "p25",
    "F25": "f25",
    "Unlimited 2024": "unlimited2024",
}


def lookup(val, data):
    val = val.replace("_", " ")
    return data[val] if val in data else val


@dataclass
class ManDetails:
    name: str
    id: int
    k: float
    entry: Literal["UPWIND", "DOWNWIND", "CROSS"]

    @staticmethod
    def parse(m, i: int):
        _m = m[0] if isinstance(m, list) else m

        return ManDetails(
            _m["info"]["short_name"],
            i + 1,
            _m["info"]["k"],
            Direction.parse(_m["info"]["start"]["direction"]),
        )


@dataclass
class ScheduleInfo:
    category: str
    name: str

    @staticmethod
    def from_str(fname):
        info = fname.split('.')[0].split("_")
        if len(info) == 1:
            return ScheduleInfo("f3a", info[0].lower())
        else:
            return ScheduleInfo(info[0].lower(), info[1].lower())

    def __str__(self):
        return f"{self.category}_{self.name}".lower()

    @staticmethod
    def lookupCategory(category):
        return lookup(category, fcj_categories)

    @staticmethod
    def lookupSchedule(schedule):
        return lookup(schedule, fcj_schedules)

    @staticmethod
    def mixed():
        return ScheduleInfo("na", "mixed")

    def fcj_to_pfc(self):
        return ScheduleInfo(
            lookup(self.category, fcj_categories), lookup(self.name, fcj_schedules)
        )

    def pfc_to_fcj(self):
        def rev_lookup(val, data):
            return (
                next(k for k, v in data.items() if v == val)
                if val in data.values()
                else val
            )

        return ScheduleInfo(
            rev_lookup(self.category, fcj_categories),
            rev_lookup(self.name, fcj_schedules),
        )

    @staticmethod
    def from_fcj_sch(sch):
        return ScheduleInfo(*sch).fcj_to_pfc()

    def to_fcj_sch(self):
        return list(self.pfc_to_fcj().__dict__.values())

    @staticmethod
    def build(category, name):
        return ScheduleInfo(category.lower(), name.lower())

    def manoeuvre_details(self) -> list[ManDetails]:
        mds = []

        for i, (k, v) in enumerate(self.json_data().items()):
            if isinstance(v, list):
                v = v[0]
            mds.append(
                ManDetails(
                    v["info"]["short_name"],
                    i + 1,
                    v["info"]["k"],
                    v["info"]["start"]["direction"],
                )
            )
        return mds

    def __eq__(self, other: ScheduleInfo):
        return str(self.fcj_to_pfc()) == str(other.fcj_to_pfc())



from __future__ import annotations

import pandas as pd
from pydantic import BaseModel

from schemas import fcj, ScheduleInfo, Direction


class MA(BaseModel):
    name: str
    id: int
    schedule: ScheduleInfo
    schedule_direction: Direction | None
    flown: list[dict] | dict

    history: dict[str, fcj.ManResult] | None = None

    mdef: dict | None = None
    manoeuvre: dict | None = None
    template: list[dict] | dict | None = None
    corrected: dict | None = None
    corrected_template: list[dict] | dict | None = None
    scores: dict | None = None

    def basic(self):
        return MA(
            name=self.name,
            id=self.id,
            schedule=self.schedule,
            schedule_direction=self.schedule_direction,
            flown=self.flown,
            history=self.history,
        )

    def simplify_history(self):
        vnames = [v[1:] if v.startswith("v") else v for v in self.history.keys()]
        vnames_old = vnames[::-1]
        vnids = [
            len(vnames) - vnames_old.index(vn) - 1
            for vn in list(pd.Series(vnames).unique())
        ]

        return MA(
            **(
                self.__dict__
                | dict(
                    history={vnames[i]: list(self.history.values())[i] for i in vnids}
                )
            )
        )


#        vids = [vnames.rindex(vn) for vn in set(vnames)]



from __future__ import annotations

import datetime
import re

import pandas as pd
from pydantic import BaseModel
from typing import Annotated


class FCJ(BaseModel):
    version: str
    comments: str
    name: str
    view: View
    parameters: Parameters
    scored: bool
    scores: list[float]
    human_scores: list[HumanResult] = []
    fcs_scores: list[Result] = []
    mans: list[Man]
    data: list[Data]
    jhash: int | None = None

    def score_df(self):
        return pd.concat(
            {fcjr.fa_version: fcjr.to_df() for fcjr in self.fcs_scores},
            axis=0,
            names=["version", "manoeuvre", "difficulty", "truncate"],
        )

    def man_df(self):
        return pd.DataFrame(
            [man.__dict__ for man in self.mans[1:-1]],
            index=pd.Index(range(len(self.mans[1:-1])), name="manoeuvre"),
        )

    def pfc_version_df(self):
        sdf = self.score_df().loc[pd.IndexSlice[:, :, 3, False]]
        return pd.concat(
            [sdf, sdf.mul(self.man_df().k, axis=0)], axis=1, keys=["raw", "kfac"]
        )

    def version_summary_df(self):
        return self.pfc_version_df().groupby("version").kfac.sum()

    def latest_version(self):
        return max([fcjr.fa_version for fcjr in self.fcs_scores])

    @property
    def id(self):
        return re.search(r"\d{8}", self.name)[0]

    @property
    def created(self):
        try:
            return datetime.datetime.strptime(
                re.search(r"_\d{2}_\d{2}_\d{2}_", self.name)[0], "_%y_%m_%d_"
            )
        except Exception:
            return None


class View(BaseModel):
    position: dict
    target: dict


class Parameters(BaseModel):
    rotation: float
    start: int
    stop: int
    moveEast: float
    moveNorth: float
    wingspan: float
    modelwingspan: float
    elevate: float
    originLat: float
    originLng: float
    originAlt: float
    pilotLat: str | float
    pilotLng: str | float
    pilotAlt: str | float
    centerLat: str | float
    centerLng: str | float
    centerAlt: str | float
    schedule: list[str]


class HumanResult(BaseModel):
    name: str
    date: datetime.date
    scores: list[float]


class Result(BaseModel):
    fa_version: str
    manresults: list[ManResult | None]

    def to_df(self) -> pd.DataFrame:
        return pd.concat(
            {i: fcjmr.to_df() for i, fcjmr in enumerate(self.manresults[1:]) if fcjmr},
            axis=0,
            names=["manoeuvre", "difficulty", "truncate"],
        )


class ManResult(BaseModel):
    els: list[El]
    results: list[Score] = []

    def to_df(self) -> pd.DataFrame:
        return pd.DataFrame(
            data=[res.score.__dict__ for res in self.results],
            index=pd.MultiIndex.from_frame(
                pd.DataFrame([res.properties.__dict__ for res in self.results])
            ),
        )

    def get_score(self, props: ScoreProperties):
        for r in self.results:
            if r.properties == props:
                return r.score


class El(BaseModel):
    name: str
    start: int
    stop: int


class Score(BaseModel):
    score: ScoreValues
    properties: ScoreProperties


class ScoreValues(BaseModel):
    intra: float
    inter: float
    positioning: float
    total: float


class ScoreProperties(BaseModel):
    difficulty: int = 3
    truncate: bool = False

    def __eq__(self, other):
        if not isinstance(other, ScoreProperties):
            return False
        return self.difficulty == other.difficulty and self.truncate == other.truncate


class Man(BaseModel):
    name: str
    k: float
    id: str
    sp: int
    wd: float
    start: int
    stop: int
    sel: bool
    background: str


class Data(BaseModel):
    VN: float = None
    VE: float = None
    VD: float = None
    dPD: float = None  #
    r: float
    p: float
    yw: float
    N: float
    E: float
    D: float
    time: int
    roll: float
    pitch: float
    yaw: float


def get_scores(file: str) -> pd.DataFrame:
    fcj = FCJ.model_validate_json(open(file, "r").read())
    return fcj.pfc_version_df()


class Origin(BaseModel):
    lat: Annotated[float, "latitude in degrees"]
    lng: Annotated[float, "latitude in degrees"]
    alt: Annotated[float, "height AMSL in meters"]
    heading: Annotated[float, "heading (direction of box Y axis) in degrees from North"] 
    move_north: Annotated[float, "non-standard offset of box"]=0
    move_east: Annotated[float, "non-standard offset of box"]=0
    move_down: Annotated[float, "non-standard offset of box"]=0
    



import numpy as np
from pydantic import BaseModel
from typing import Annotated
from .utils.enum import EnumStr
import numpy as np


class MBTags:
    CENTRE = 0


def centred(elb):
    setattr(elb, "centred", True)
    return elb

c45 = np.cos(np.radians(45))


def r(turns):
    return (2 * np.pi * np.array(turns)).tolist()

class Orientation(EnumStr):
    UPRIGHT = np.pi
    INVERTED = 0


class Heading(EnumStr):
    RTOL = np.pi
    LTOR = 0
    OUTTOIN = 3 * np.pi / 2
    INTOOUT = np.pi / 2

    @staticmethod
    def values():
        return np.array(list(Heading.__members__.values()))

    @staticmethod
    def infer(bearing: Annotated[float, "in radians from north"]):
        def check(bearing: float, heading: Heading):
            return (
                np.round(np.abs(4 * (bearing - heading.value)) / (2 * np.pi)).astype(
                    int
                )
                % 4
            ) == 0

        for head in Heading.__members__.values():
            if check(bearing, head):
                return head
        else:
            raise ValueError(f"Invalid bearing {bearing}")

    def reverse(self):
        return {
            Heading.RTOL: Heading.LTOR,
            Heading.LTOR: Heading.RTOL,
            Heading.OUTTOIN: Heading.INTOOUT,
            Heading.INTOOUT: Heading.OUTTOIN,
        }[self]


class Direction(EnumStr):
    UPWIND = 1
    DOWNWIND = -1
    CROSS = 0

    def wind_swap_heading(self, d_or_w: Heading) -> int:
        match self:
            case Direction.UPWIND:
                return d_or_w
            case Direction.DOWNWIND:
                return d_or_w.reverse()
            case Direction.CROSS:
                return d_or_w

    @staticmethod
    def parse(s: str):
        match s[0].lower():
            case "u":
                return Direction.UPWIND
            case "d":
                return Direction.DOWNWIND
            case "c":
                return Direction.CROSS
            case _:
                raise ValueError(f"Invalid wind {s}")


class Height(EnumStr):
    BTM = 0.2
    MID = 0.6
    TOP = 1.0


class Position(EnumStr):
    CENTRE = 0
    END = 1


class BoxLocation(BaseModel):
    height: Height
    direction: Direction | None = None
    orientation: Orientation | None = None


def boxlocationmaker(
    height: Height,
    direction: Direction | None = None,
    orientation: Orientation | None = None,
):
    return BoxLocation(
        height=height,
        direction=direction,
        orientation=orientation,
    )



from pytest import fixture, mark
from schemas import MA
from pathlib import Path
from json import load

@fixture(scope="session")
def ma_dict():
    return load(Path(__file__  + "/data/flight.ajson").open())['mans'][0]

@mark.skip(reason="no ")
def test_ma_validate(ma_dict):
    man = MA.model_validate(ma_dict)
    assert isinstance(man, MA)


from .positioning import Position, BoxLocation
from typing import Tuple, Annotated
from pydantic import BaseModel


class ManInfo(BaseModel):
    name: str
    short_name: str
    k: float
    position: Position
    start: BoxLocation
    end: BoxLocation
    centre_points: Annotated[
        list[int],
        "points that should be centered, ids correspond to the previous element",
    ] = []
    centred_els: Annotated[
        list[Tuple[int, float]], "element ids that should be centered"
    ] = []

    def to_dict(self):
        return self.model_dump()
    
    @staticmethod
    def from_dict(data: dict):
        return ManInfo.model_validate(data)


def maninfomaker(
    name: str,
    short_name: str,
    k: float,
    position: Position,
    start: BoxLocation,
    end: BoxLocation,
    centre_points: Annotated[
        list[int],
        "points that should be centered, ids correspond to the previous element",
    ] = None,
    centred_els: Annotated[
        list[Tuple[int, float]], "element ids that should be centered"
    ] = None,
):
    return ManInfo(
        name=name,
        short_name=short_name,
        k=k,
        position=position,
        start=start,
        end=end,
        centre_points=centre_points if centre_points is not None else [],
        centred_els=centred_els if centred_els is not None else [],
    )



from schemas.positioning import BoxLocation, Height, Direction


def test_enumstr():
    h = Height.MID
    assert h == Height.MID
    assert h.value == 0.6


def test_boxloc_serialize():
    bloc = BoxLocation(height=Height.MID, direction=Direction.UPWIND, orientation=None)

    bloc_dict = bloc.model_dump()

    assert bloc_dict["height"] == "MID"
    assert bloc_dict["direction"] == "UPWIND"
    assert bloc_dict["orientation"] is None



    bloc_parsed = BoxLocation.model_validate_json(bloc.model_dump_json())

    assert bloc == bloc_parsed
